#!/usr/bin/python3
# -*- coding: utf-8 -*-
import apt
import dbus
import dbus.service
import dbus.mainloop.glib
import os
import time
import datetime
import shutil
import psutil
import threading
import re
import apt_pkg
import gc
import logging
import subprocess#snap


from apt.debfile import DebPackage
from apt.progress.base import AcquireProgress
from kylindbus.apt_handler import AppActions, AptHandler, WorkitemError
from kylindbus.enums import INTERFACE, INTERFACES, check_apt_lock
from kylindbus.policykit_service import PolicyKitService
from kylindbus.log_manager import logger
from kylindbus.sqlite3_server import Sqlite3Server
from kylindbus.log_message_server import ConnectServerThread
from kylindbus.parse_json import ParseJson

CRONPATH = "/var/spool/anacron/"
SOURCESPATH = "/etc/apt/sources.list.d"
CODEHOME = "/usr/share/kylin-update-manager/"
PACKAGES_MANAGER_TOOLS = 'cn.kylinos.KylinUpdateManager.action'



class WorkItem:
    def __init__(self, pkgname, action, kwargs):
        self.pkgname = pkgname
        self.action = action
        self.kwargs = kwargs


class WorkThread(threading.Thread):
    def __init__(self, dbusDaemon):
        threading.Thread.__init__(self)
        self.dbusDaemon = dbusDaemon
        ("KUM DBUS START ", type(self.dbusDaemon))

    #
    # 函数:线程执行函数
    #
    def run(self):
        while True:
            if len(self.dbusDaemon.worklist) == 0:
                time.sleep(0.2)
                continue
            self.dbusDaemon.mutex.acquire()
            self.item = self.dbusDaemon.worklist.pop(0)
            self.dbusDaemon.mutex.release()
            try:
                func = getattr(self.dbusDaemon.aptHandler, self.item.action)
                if func is None:
                    print("Error action: ", self.item)
                res = func(self.item.pkgname, self.item.kwargs)
                if res is False:
                    print("Action exec failed...")
            except WorkitemError as e:
                # 如果是包列表，此时强制装换为str，后面显示str将出现异常，故而暂不显示包列表
                kwarg = {"apt_appname": str(self.item.pkgname),
                         "apt_percent": str(e.errornum),
                         "error_message": str(e.details),
                         }
                # 安装失败时需要进行二次判断，防止第一次安装失败导致后续的包都安装失败
                result = ""
                results = os.popen("dpkg -l | grep %s" %
                                   self.item.pkgname).readlines()
                for res in results:
                    res = re.sub(" +", " ", res)
                    if str(res).split(" ")[1] == self.item.pkgname:
                        result = res.split(" ")[0]
                if result != "ii":
                    logger.error("安装状态：失败 - 包名：" + str(self.item.pkgname) +
                                 " - 错误码：(" + str(e.errornum) + ") - 详情：" + str(e.details))
                    self.dbusDaemon.kum_apt_signal("apt_error", kwarg)
                    self.dbusDaemon.insert_app_message("apt_error", kwarg)
            except Exception as e:  # 代码异常
                kwarg = {"apt_appname": str(self.item.pkgname),
                         "apt_percent": '24',
                         "error_message": str(e),
                         }
                logger.error("安装状态：失败 - 包名：" + str(self.item.pkgname) +
                             " - 错误码：(24) - 详情：" + str(e), exc_info=True)
                # self.dbusDaemon.kum_apt_signal("apt_error", kwarg)
                # self.dbusDaemon.insert_app_message("apt_error", kwarg)
            finally:
                self.dbusDaemon.error_message = ""
                self.item = None
                self.dbusDaemon.counter -= 1
            time.sleep(0.3)
            self.dbusDaemon.aptHandler.delete_cache()


# acquire progress
class FetchProgress(AcquireProgress):
    def __init__(self, Daemon) -> None:
        super().__init__()
        self.deamon = Daemon

    def pulse(self, owner):
        # partial_present = owner.partial_present
        # total_needed = owner.total_needed
        progress = int((self.current_bytes / self.total_bytes) * 100)
        total_size = self.total_bytes
        # progress = int((partial_present/total_needed)*100)
        # total_size = float(total_needed/(1024**2))
        logging.info("fetch progress:%d,total size:%f" % (progress, total_size))
        self.deamon.DownloadProgressSignal(progress, total_size)
        return

    def stop(self):
        logging.info("fetch progess finished")
        finish_status = False
        if self.current_items == self.total_items:
            finish_status = True
        else:
            finish_status = False
        self.deamon.DownloadFinishSignal(finish_status)
        return

    def fail(self, acquireitemdesc):
        desc = acquireitemdesc.shortdesc
        err_text = acquireitemdesc.owner.error_text
        status = acquireitemdesc.owner.status
        type = ""
        if status == apt_pkg.AcquireItem.STAT_AUTH_ERROR:
            type = "authentication error"
        elif status == apt_pkg.AcquireItem.STAT_ERROR:
            type = "fetch error"
        elif status == apt_pkg.AcquireItem.STAT_TRANSIENT_NETWORK_ERROR:
            type = "network error"
        elif status == apt_pkg.AcquireItem.STAT_IDLE:
            type = "to be fetched"
        elif status == apt_pkg.AcquireItem.STAT_FETCHING:
            type = "fetching"
        elif status == apt_pkg.AcquireItem.STAT_DONE:
            type = "finished"
        else:
            type = "normal"
        self.deamon.DownloadErrorSignal(desc, type, err_text)
        return


class Daemon(PolicyKitService):
    def __init__(self, bus, mainloop):
        self.aptHandler = AptHandler(self)
        self.parse_json = ParseJson(self)
        self.sqlite_server = Sqlite3Server()
        self.worklist = []
        self.cancel_name_list = []
        self.mutex = threading.RLock()
        self.cancelmutex = threading.RLock()
        self.worker_thread = WorkThread(self)
        self.worker_thread.setDaemon(True)
        self.worker_thread.start()
        self.importantwork = False
        self.efi_status = ""
        self.optional_message_list = []
        self.silent_message_list = []
        self.bus_name = dbus.service.BusName(INTERFACE, bus=bus)
        PolicyKitService.__init__(self, self.bus_name, INTERFACES)
        self.counter = 0
        self.mainloop = mainloop
        self.error_message = ""
        self.autoupdate_status = 0
        self.json_list = []
    #
    # 函数：检测计数器是否为0，是否可以退出dbus
    #
    def check_counter(self):
        if self.counter == 0:
            logger.info("自我升级完成，定时自动重启服务")
            self.mainloop.quit()
        else:
            exit_timer = threading.Timer(5, self.check_counter)
            exit_timer.start()

    #
    # 函数：apt安装结果写入数据库
    #
    def insert_app_message(self, type_msg, msg):
        try:
            self.insert_to_sqlite(type_msg, msg, self.optional_message_list, 1)
            self.insert_to_sqlite(type_msg, msg, self.silent_message_list, 0)
        except Exception as e:  # 数据库文件被删除或者新增字段导致需要重新初始化数据库再写入
            self.sqlite_server.init_sqlit()
            self.insert_to_sqlite(type_msg, msg, self.optional_message_list, 1)
            self.insert_to_sqlite(type_msg, msg, self.silent_message_list, 0)
            logger.error("数据库写入异常，包名：%s" % msg["apt_appname"], exc_info=True)

    #
    # 函数：插入app信息到数据库并发送信号
    #
    def insert_to_sqlite(self, type_msg, msg, list, var=1):
        time = datetime.datetime.now()
        timestr = datetime.datetime.strftime(time, "%Y-%m-%d %H:%M:%S")
        n = 0
        for i in list:
            if i["appname"] == msg["apt_appname"]:
                appmsg = list.pop(n)
                icon = None
                if type_msg == "apt_finish":
                    statue = "Success"
                    description = appmsg["description"]
                else:
                    statue = "Fail"
                    if msg.get("error_message") is not None:
                        description = msg["error_message"]
                    else:
                        description = self.error_message

                if msg["apt_percent"] == "100":
                    errorcode = "0"
                else:
                    errorcode = msg["apt_percent"]
                if var == 1:
                    server_errorcode = 10010000 + int(errorcode)
                else:
                    server_errorcode = 10000000 + int(errorcode)
                self.sqlite_server.insert_into_installed(appmsg["appname"], appmsg["source_version"],
                                                         timestr, description, icon,
                                                         statue, i["keyword"], errorcode)
                self.sqlite_server.insert_into_display(
                    'update_time', timestr)  # 写入本地数据库
                if i["keyword"] == "1":  # 可选更新才发送信号用于更新历史界面
                    self.update_sqlite_signal(appmsg["appname"], timestr)
                self.send_message_to_server(
                    appmsg["appname"], appmsg["source_version"], statue, str(server_errorcode))  # 发送到日志服务器
                logger.info("安装状态成功写入数据库，包名：%s" % msg["apt_appname"])
            n += 1

    #
    # 函数：插入deb本地包安装信息到数据库
    #
    def insert_deb_message(self, kum, msg):
        time = datetime.datetime.now()
        timestr = datetime.datetime.strftime(time, "%Y-%m-%d %H:%M:%S")
        if msg["statue"] == "Fail":
            description = msg["error_message"]
        else:
            description = kum["description"]
        try:
            self.sqlite_server.insert_into_installed(kum["appname"], kum["version"],
                                                     timestr, description, kum["icon"],
                                                     msg["statue"], kum["keyword"], msg["errorcode"])
        except:
            self.sqlite_server.init_sqlit()
            self.sqlite_server.insert_into_installed(kum["appname"], kum["version"],
                                                     timestr, description, kum["icon"],
                                                     msg["statue"], kum["keyword"], msg["errorcode"])

    #
    # 函数:新增任务
    # item：任务
    #
    def add_worker_item(self, item):
        print("####add_worker_item:", item)
        self.mutex.acquire()
        self.worklist.append(item)
        self.mutex.release()

    #
    # 函数：获取当前下载包的依赖
    #
    def get_deps_pkg(self, appname):
        url_list = []
        name_list = []
        size_list = []
        fullname_list = []
        all_size = 0
        depends_status = False
        try:
            res = os.popen("LANG= apt-get install --print-uris %s -y" %
                           appname).readlines()
            # print(len(res))
            num = 0
            for i in res:
                if "After this operation," in i:
                    num = num + 1
                    depends_status = True
                    while num < len(res):
                        i = res[num]
                        i_list = i.strip("\n").split(" ")
                        url_list.append(i_list[0].strip("'"))
                        name_list.append(i_list[1].split("_")[0])
                        fullname_list.append(i_list[1])
                        size_list.append(i_list[2])
                        all_size = all_size + int(i_list[2])
                        num = num + 1
                    break
                # 软件包已安装(kylin-update-desktop单独处理)-lj
                elif "0 upgraded," in i and " 0 newly" in i:
                    # print("len:",len(res), "num:", num+1)
                    if len(res) == num + 1:
                        depends_status = True
                        break
                elif "The following packages have unmet dependencies:" in i:
                    logger.error("--%s获取依赖失败--" %
                                 appname + "\n" + "".join(res))
                    depends_status = False
                    break
                num = num + 1
            return url_list, name_list, fullname_list, size_list, all_size, depends_status
        except Exception as e:
            logger.error("依赖解析代码出错--%s--" % appname, exc_info=True)
            return url_list, name_list, fullname_list, size_list, all_size, depends_status

    #
    # 函数：判断efi分区的状态
    #
    def get_efi_status(self):
        with open("/proc/mounts") as f:
            for line in f.readlines():
                line_section = line.split(" ")
                if "/boot/efi" in line_section[1]:
                    status = line_section[3].split(",")
                    for i in status:
                        if i == "ro" or i == "rw":
                            return i
                    break

    #
    # 函数：安装前后设置efi分区的状态为读写
    #
    def set_efi_status(self, before=True):
        try:
            if self.efi_status == "ro" and before == True:
                os.system("mount -o remount,rw /boot/efi")
                os.system("mount -o remount,rw /boot")
            elif self.efi_status == "ro" and before == False:
                os.system("mount -o remount,ro /boot/efi")
                os.system("mount -o remount,ro /boot")
        except Exception as e:
            logger.error("修改/boot/efi挂载状态出错", exc_info=True)

    #
    # 函数：开启线程收集客户端信息发送给php服务器
    #
    def send_message_to_server(self, _appname="", _appversion="", _statue="", _errorcode=""):
        send_thread = ConnectServerThread(self)
        send_thread.get_values(_appname, _appversion, _statue, _errorcode)
        send_thread.start()

    #
    # 函数：共用函数用于设置apt状态并发送信号。
    #
    def send_apt_status(self, name, percent, message, status=True):
        if status:
            kwarg = {"apt_appname": str(name), "apt_percent": str(
                percent), "action": str(message)}
            self.kum_apt_signal("apt_finish", kwarg)
        else:
            kwarg = {"apt_appname": str(name), "apt_percent": str(
                percent), "error_message": str(message)}
            self.kum_apt_signal("apt_error", kwarg)
        self.set_efi_status(False)
        self.counter -= 1

    def send_deb_status(self, name, percent, message, status=True):
        if status:
            kwarg = {"apt_appname": str(name), "apt_percent": str(
                percent), "action": str(message)}
            self.kum_deb_signal("apt_finish", kwarg)
        else:
            kwarg = {"apt_appname": str(name), "apt_percent": str(
                percent), "error_message": str(message)}
            self.kum_deb_signal("apt_error", kwarg)
        self.set_efi_status(False)
        self.counter -= 1
    #
    #函数：设置snap包的安装状态并发送信号。
    #
    def send_snap_status(self, message, status=True):
        if status:
            logger.info("安装成功----------------")
            kwarg = {"action": str(message)}
            self.kum_snap_signal("snap_finish", kwarg)
        else:
            logger.info("安装失败-------------------------")
            kwarg = {"error_message": str(message)}
            self.kum_snap_signal("snap_error", kwarg)
        self.set_efi_status(False)
        self.counter -= 1
    #
    # 函数:初始化cache
    #
    def init_cache(self):
        self.aptHandler.init_cache()

    #
    # 函数:提权
    # polkit()
    #
    def auth_with_policykit(self, lang, text, get_action_id, sender=None):
        # return True
        action = get_action_id
        # lang = os.getenv('LANG')
        if text == "":
            if "zh_CN" in lang:
                text = "要卸载软件"
            else:
                text = "To uninstall the software"
        if not sender:
            raise ValueError('sender == None')
        print("text----",text)
        #        print "auth_with_policykit:", sender
        granted = False
        try:
            obj = dbus.SystemBus().get_object('org.freedesktop.PolicyKit1',
                                              '/org/freedesktop/PolicyKit1/Authority')
            policykit = dbus.Interface(obj, 'org.freedesktop.PolicyKit1.Authority')

            subject = ('system-bus-name', {'name': sender})
            flags = dbus.UInt32(1)  # AllowUserInteraction flag
            if "zh_CN" in lang:
                msg = text + "，您需要验证。"
            else:
                msg = text + "，You need to verify。"
            details = {'polkit.message': msg}
            cancel_id = ''  # No cancellation id
            (granted, notused, details) = policykit.CheckAuthorization(
                subject, action, details, flags, cancel_id)
        except Exception as e:
            print("auth with except......", str(e))
            granted = False

        return granted

    def get_proc_from_dbus_name(self,dbus_name, bus=None):
        """Return a deferred that gets the id of process owning the given
        system D-Bus name.
        """

        if not bus:
            bus = dbus.SystemBus()
        bus_obj = bus.get_object("org.freedesktop.DBus",
                                "/org/freedesktop/DBus/Bus")
        pid = bus_obj.GetConnectionUnixProcessID(dbus_name,
                                        dbus_interface="org.freedesktop.DBus")

        try:
            absolute_path = os.readlink(f"/proc/{int(pid)}/exe")
        except FileNotFoundError:
            absolute_path = ""

        return absolute_path
    # -------------------------------------dbus methods----------------------------------------

    #
    # 函数:安装和升级
    # install package sa:kum_fetch_signal() and kum_apt_signal()
    #
    @dbus.service.method(INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    def install_and_upgrade(self, pkgName, sender=None):
        allow_fix_env = ["/usr/bin/kylin-printer","/usr/bin/kylin-installer"]

        sender_name = self.get_proc_from_dbus_name(sender)
        logger.debug("####install_and_upgrade: pkgName:%s,sender_name:%s",pkgName,sender_name)

        if sender_name not in allow_fix_env:
            raise dbus.exceptions.DBusException("ERROR: You are not allowed to perform this action.")
        
        if ";" in pkgName or " " in pkgName:
            return False
        logger.debug("####install_and_upgrade: ", pkgName)
        self.counter += 1
        item = WorkItem(pkgName, AppActions.INS_AND_UP, None)
        self.add_worker_item(item)

        logger.debug("####install_and_upgrade return")
        return True
    #
    # 函数：判断如果升级完成，发送一个信号
    #
    @dbus.service.method(INTERFACE)
    def send_signal_upgrade(self):
        logger.info("发送信号----------")
        def check_process(process_name):
            for proc in psutil.process_iter(['pid', 'name']):
                if proc.info['name'] == process_name:
                    return True
            return False
        
        kwarg = {"apt_appname": str("kylin-update-desktop-support"),
                 "apt_percent": str(100),
                 "error_message": str("no"),
                 }
        #fdbj = open("/tmp/kylin-update.lock", 'w')
        #fd = fdbj.fileno()
        #try:
            # 加锁成功
        #    fcntl.flock(fd, fcntl.LOCK_EX|fcntl.LOCK_NB)
        #except:
            #加锁失败，说明正在更新
        #      r = os.popen("dpkg -l |grep 'iF|iU'")
        #      text = r.read()
        #      r.close()
        #      if text == "":
        self.kum_apt_signal("apt_finish", kwarg)
        os.system("rm /tmp/lock/kylin-update.lock")
        logger.info("执行提示语")
        
        try:
            if check_process("ukui-kwin_wayland"):
                os.environ["XDG_RUNTIME_DIR"] = "/run/user/1000/"
                os.environ["QT_QPA_PLATFORM"] = "wayland"
        except Exception as e:
            logger.error(str(e))

        os.environ["DISPLAY"] = ":0"
        p = subprocess.run("/usr/sbin/untitled1", stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)


    # -------------------------------------dbus signals----------------------------------------
    #
    # dbus信号：用于发送apt下载时进度信息
    #
    @dbus.service.signal(INTERFACE, signature='sa{sv}')
    def kum_fetch_signal(self, type_msg, msg):
        pass

    #
    # dbus信号：用于发送apt安装时进度信息
    #
    @dbus.service.signal(INTERFACE, signature='sa{sv}')
    def kum_apt_signal(self, type_msg, msg):
        pass

    #
    # dbus 信号：用于发送本地安装卸载进度信息（不与更新混用）
    #
    @dbus.service.signal(INTERFACE, signature='sa{sv}')
    def kum_deb_signal(self, type_msg, msg):
        pass

    #
    # dbus 信号：用于发送本地安装snap包的结果
    #
    @dbus.service.signal(INTERFACE, signature='sa{sv}')
    def kum_snap_signal(self, type_msg, msg):
        pass

    #
    # dbus 信号：用于发送卸载中信号
    #
    @dbus.service.signal(INTERFACE, signature='')
    def kum_deb_remove(self):
        pass

    #
    # dbus信号：用于拷贝包到/var/cache/apt/archives的完成信号
    #
    @dbus.service.signal(INTERFACE, signature="s")
    def copy_finish(self, appname):
        pass

    #
    # dbus信号：发送重要更新包列表信息
    #
    @dbus.service.signal(INTERFACE, signature='a{sv}asasasassb')
    def important_app_message_signal(self, msg, url_list, name_list, size_list, fullname_list, all_size,
                                     depends_status):
        pass

    #
    # dbus信号：发送关键更新包列表信息
    #
    @dbus.service.signal(INTERFACE, signature='a{sv}')
    def crucial_app_message_signal(self, msg):
        pass

    #
    # dbus信号：通知前端获取重要更新信息获取完成的dbus信号
    #
    @dbus.service.signal(INTERFACE, signature='s')
    def get_message_finished_signal(self, number):
        pass

    #
    # dbus信号：通知前端数据库写入完成后的dbus信号，用于实时刷新历史更新界面
    #
    @dbus.service.signal(INTERFACE, signature='ss')
    def update_sqlite_signal(self, appname, time):
        pass

    # dbus signal:send the progress info of certain packages download
    @dbus.service.signal(INTERFACE, signature='ii')
    def DownloadProgressSignal(self, progress, total_size):
        pass

    # dbus signal:send the finish signal after each certain packages download
    @dbus.service.signal(INTERFACE, signature='b')
    def DownloadFinishSignal(self, status):
        pass

    # dbus signal:send the error text when there're errors during download
    @dbus.service.signal(INTERFACE, signature='sss')
    def DownloadErrorSignal(self, desc, type, err_text):
        pass
